/* Alternative checkpassword for QPopup by Jedi/Sector One <j@4u.net> 
 *
 * Format of the configuration file is :
 * pop_login:crypted_password:real_login:path
 *
 * Syslog logging added by Paul Gregg <pgregg@tibus.net> Dec 1997
 * compile by: gcc -o checkpoppasswd checkpoppasswd.c -lcrypt
 * Original URL of this file: http://www.pgregg.com/projects/
*/

#ifndef LOCALPOP_PASSWORDS
#define LOCALPOP_PASSWORDS "/var/qmail/poppasswd"
#endif

#ifndef DONTUSEQMAILGROUP
#ifndef QMAIL_GROUPNAME
#define QMAIL_GROUPNAME "qmail"
#endif
#endif

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <grp.h>
#include <sys/types.h>
#include <pwd.h>
#include <errno.h>
#include <syslog.h>

#define  BUFLEN 2048

#ifndef LINE_MAX
# define LINE_MAX 2048
#endif

#define VERBOSELOG if (verbosity>1) syslog
#define NORMALLOG if (verbosity>0) syslog

extern int errno;

extern char *crypt();
extern char *malloc();

const char *remoteip;
const char *remotehost;
const char *remoteinfo;

char verbosity=1;

struct newpw {
   char *pw_passwd;
   char *pw_dir;   
   char *pw_name;      
   char *pw_shell;         
   uid_t pw_uid;
   gid_t pw_gid;
   char *pw_real;
} *newgetpwnam(char *newlogin)
{
   static struct newpw npw;
   struct passwd *pw;
#ifndef DONTUSEPOPPWDFILE
   static char line[LINE_MAX];
   char *tlogin = NULL;
   FILE *fp;
   
   if ((fp = fopen(LOCALPOP_PASSWORDS, "rt")) == NULL) {
     NORMALLOG(LOG_AUTH | LOG_WARNING, "No file");
     _exit(2);
   }
   while (fgets(line, LINE_MAX, fp) != NULL) {
      char *linepnt;
      
      if ((linepnt = strchr(line, '\n')) != NULL) {
         *linepnt = 0;
      }
      if ((linepnt = strchr(line, ':')) == NULL) {
        NORMALLOG(LOG_AUTH | LOG_WARNING, LOCALPOP_PASSWORDS " corrupt - a line with no colons");
        _exit(2);
      }
      *linepnt++=0;
      if (strcmp(line, newlogin) == 0) {
         npw.pw_real = line;
         npw.pw_passwd = linepnt;
         if ((linepnt = strchr(linepnt, ':')) == NULL) {
           NORMALLOG(LOG_AUTH | LOG_WARNING, LOCALPOP_PASSWORDS " corrupt - line with only one colon");
           _exit(2);
         }
         *linepnt++=0;
         tlogin=linepnt;
         if ((linepnt = strchr(linepnt, ':')) == NULL) {
           NORMALLOG(LOG_AUTH | LOG_WARNING, LOCALPOP_PASSWORDS " corrupt - two-colon line.");
           _exit(2);
         }
         *linepnt++=0;
         npw.pw_dir = linepnt;
         if ((linepnt = strchr(linepnt, ':')) != NULL) {
          *linepnt=0;
         }
         break;
      }
   }   
   fclose(fp);
   
   if (tlogin == NULL) {
     NORMALLOG(LOG_AUTH | LOG_WARNING, "FAILED: %s - no such user from %s@%s [%s]", newlogin, remoteinfo, remotehost, remoteip);
     sleep(2);
     _exit(1);
   }
   pw = getpwnam(tlogin);
   if (!pw) {
     NORMALLOG(LOG_AUTH | LOG_WARNING, "getpwnam failed for %s",tlogin);
     _exit(1); /* XXX: unfortunately getpwnam() hides temporary errors */
   }
   npw.pw_name = tlogin;
#else
   pw = getpwnam(newlogin);
   if (!pw) {
     NORMALLOG(LOG_AUTH | LOG_WARNING, "FAILED: %s - probably no such user from %s@%s [%s]", newlogin, remoteinfo, remotehost, remoteip);
     sleep(2);
     _exit(1);
   }
   npw.pw_name = pw->pw_name;
   npw.pw_real = pw->pw_name;
   npw.pw_passwd = pw->pw_passwd;
   npw.pw_dir = pw->pw_dir;
#endif //DONTUSEPOPPWDFILE
      
   npw.pw_shell = pw->pw_shell;
   npw.pw_uid = pw->pw_uid;
   npw.pw_gid = pw->pw_gid;
   
   return &npw;
}

int main(int argc, char **argv, char *envp[])
{
   struct newpw *pw;
   char *login;
   char *password;
   char *stored;
   char *encrypted;
   int r;
   int i;
   char up[513];
   int uplen;

   while (argc>2 && argv[1][0]=='-') {
     argv++; argc--;
     if (argv[0][1]=='v') verbosity++;
     else if (argv[0][1]=='q') verbosity--;
   }
   if (argc < 2) {
     NORMALLOG(LOG_AUTH | LOG_WARNING, "Not enough parameters.");
     _exit(2);
   }

   remoteip=getenv("TCPREMOTEIP");
   remotehost=getenv("TCPREMOTEHOST");
   remoteinfo=getenv("TCPREMOTEINFO");

   VERBOSELOG(LOG_AUTH | LOG_WARNING, "POP3 connect from %s@%s [%s]", remoteinfo, remotehost, remoteip);



   uplen = 0;
   for (;;) {
      do {
        r = read(3,up + uplen,sizeof(up) - uplen);
      } while ((r == -1) && (errno == EINTR));
      if (r == -1) _exit(111);
      if (r == 0) break;
      uplen += r;
      if (uplen >= sizeof(up)) _exit(1);
   }
   close(3);

#define BADINPUTDATA "Bad input data. Possible overflow attempt from %s.",remoteip

   i = 0;
   login = up + i;
   while (up[i++])
     if (i == uplen) {
        NORMALLOG(LOG_AUTH | LOG_WARNING, BADINPUTDATA);
       _exit(2);
     }
   password = up + i;
   while (up[i++])
   if (i == uplen) {
     NORMALLOG(LOG_AUTH | LOG_WARNING, BADINPUTDATA);
     _exit(2);
   }
   if (i==uplen) {
     NORMALLOG(LOG_AUTH | LOG_WARNING, BADINPUTDATA);
     _exit(2);
   }
   pw = newgetpwnam(login);
   if (!pw) {
     NORMALLOG(LOG_AUTH | LOG_WARNING, BADINPUTDATA);
     _exit(1); 
   }
   stored = pw->pw_passwd;
   
   encrypted = crypt(password,stored);

   setenv("VIRTUALUSER",login,1);

   for (i = 0;i < sizeof(up);++i) up[i] = 0;

   if (!*stored || strcmp(encrypted,stored)) {
     NORMALLOG(LOG_AUTH | LOG_WARNING, "FAILED: %s [%s] - password incorrect from %s@%s [%s]", login, pw->pw_real, remoteinfo, remotehost, remoteip);
     sleep(2); /* pgregg sleep to 1)allow syslog to work and 2)hinder pop3 brute force hacks */
     _exit(1);
   }
   
   if (!pw->pw_uid) {
     NORMALLOG(LOG_AUTH | LOG_WARNING, "no uid for user %s",pw->pw_name);
     _exit(1);
   }

   {
    if (setgid(pw->pw_gid) == -1) {
     NORMALLOG(LOG_AUTH | LOG_WARNING, "couldn't set gid %d",pw->pw_gid); 
     _exit(1);
    }
#ifndef DONTUSEQMAILGROUP
    {
     struct group *pqmailgroup=getgrnam(QMAIL_GROUPNAME);
     if (!pqmailgroup || setegid(pqmailgroup->gr_gid)==-1) {
      NORMALLOG(LOG_AUTH | LOG_WARNING, "couldn't set qmail gid"); 
     }
    }
#endif
   }
   if (setuid(pw->pw_uid) == -1) {
     NORMALLOG(LOG_AUTH | LOG_WARNING, "couldn't set uid %d",pw->pw_uid);
     _exit(1);
   }
   if (chdir(pw->pw_dir) == -1) {
     NORMALLOG(LOG_AUTH | LOG_WARNING, "FAILED: %s - unable to change to Mail directory: %s", pw->pw_real, pw->pw_dir);
     _exit(111);
   }

   if (getuid() < 500) {
     NORMALLOG(LOG_AUTH | LOG_WARNING, "DENIED: POP3 user's UID < 500");
     _exit(1);
   }

  setenv("USER",pw->pw_name,1);
  setenv("HOME",pw->pw_dir,1);
  setenv("SHELL",pw->pw_shell,1);

  VERBOSELOG(LOG_AUTH | LOG_WARNING, "POP3 user %s : %s logged in from %s@%s [%s]", pw->pw_real, pw->pw_dir, remoteinfo, remotehost, remoteip);
  execvp(argv[1],argv + 1);
  _exit(111);

  return 0;
}

